#!/usr/bin/php
<?php
// Copyright (C) 2000-2017 Synology Inc. All rights reserved.

define ("RET_KEY_ERR_MSG", "err_msg");
define ("RET_KEY_RESULT", "result");
define ("RET_KEY_DATA_VERSION", "app_data_version");
define ("DATA_VERSION", "2.0");

define ("ERR_MSG_SECTION", "bkp");
define ("ERR_MSG_KEY_BKP_FAIL", "bkp_backup_fail");
define ("ERR_MSG_KEY_OP_FAIL", "bkp_op_fail");
define ("ERR_MSG_KEY_RESTORE_DATA_TOO_NEW", "bkp_restore_data_too_new");
define ("ERR_MSG_KEY_RESTORE_FAIL", "bkp_restore_fail");
define ("ERR_MSG_KEY_WRONG_PARAM", "bkp_wrong_param");

$blCancel = false;

$DangerWord = array (
	'\\',
	'`'
);

declare(ticks = 1);

function sig_handler ($signo)
{
	global $blCancel;
	switch ($signo) {
		case SIGTERM:
			$blCancel = true;
			break;
		default:
	}
}

pcntl_signal(SIGTERM, "sig_handler");


/**
 * Set error message by key and write syslog if has $logMsg
 *
 * @param logMsg [IN] Message to write syslog
 * @param errMsg [OUT] Record error msg
 * @param key [IN] Key
 *
 */
function SetErrMsg ($logMsg, $errMsg, $key = ERR_MSG_KEY_WRONG_PARAM)
{
	$errMsg->{"section"} = ERR_MSG_SECTION;
	$errMsg->{"key"} = $key;
	if ("" !== $logMsg) {
		syslog(LOG_ERR, $logMsg);
	}
}

function exec_get_stderr($cmd, &$stderr=null) {
	$proc = proc_open($cmd, [
		2 => ['pipe','w'],
	], $pipes);
	$stderr = stream_get_contents($pipes[2]);
	fclose($pipes[2]);
	return proc_close($proc);
}

/**
 * Check if MariaDB is enabled?
 *
 * @return true: enabled
 *        false: disabled
 */
function IsMariaDBEnabled ()
{
	$ret = false;
	$execOutput = array();
	$execStatus = -1;

	exec("/var/packages/MariaDB/scripts/start-stop-status status >/dev/null 2>&1", $execOutput, $execStatus);
	if (0 == $execStatus) {
		$ret = true;
	}

	return $ret;
}

/**
 * Wait MariaDB to be enabled?
 *
 * @return 0: MariaDB get enabled
 *        -1: MariaDB doesn't get enabled after MAX_WAIT_TIME waiting time
 */
function MariaDBWaitForEnabled ()
{
	$MAX_WAIT_TIME = 20;
	$ret = -1;

	for ($i = 0; $i < $MAX_WAIT_TIME; $i++) {
		if (!IsMariaDBEnabled()) {
			sleep(1);
			continue;
		}
		$ret = 0;
		break;
	}

	return $ret;
}

/**
 * Check self argument
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 * @param bkpOutput [IN] Backup framework output argument, write output into this file path.
 *                    Should not be ""
 * @param type [IN] 0: export db, 1: import db
 * @param errMsg [OUT] Record error msg
 *
 * @return 0: success
 *        -1: fail
 */
function SelfArgCheck ($bkpInput, $bkpOutout, $type, $errMsg)
{
	$ret = -1;
	$DATA_VERSION = floatval(DATA_VERSION);

	if (!property_exists($bkpInput, 'temp_path')) {
		SetErrMsg("MariaDB_SelfArgCheck: temp_path: no param!", $errMsg);
		goto End;
	}

	if ("" === $bkpInput->{'temp_path'}) {
		SetErrMsg("MariaDB_SelfArgCheck: temp_path: must have value!", $errMsg);
		goto End;
	}

	if (0 == $type) {
		$ret = 0;
		goto End;
	}

	//for 5.1 beta bakcup framework
	if (property_exists($bkpInput, 'old_mysql_backup')) {
		$ret = 0;
		goto End;
	}

	if (!property_exists($bkpInput, 'app_data_version')) {
		SetErrMsg("MariaDB_SelfArgCheck: app_data_version: no param!", $errMsg);
		goto End;
	}

	if (preg_match('/^\d+$/', $bkpInput->{'app_data_version'})) {
		if (0 != floatval($bkpInput->{'app_data_version'})) {
			SetErrMsg("MariaDB_SelfArgCheck: app_data_version: must be format %d.%d or 0", $errMsg);
			goto End;
		}
		$ret = 0;
		goto End;
	}

	if (!preg_match('/^\d+\.\d+$/', $bkpInput->{'app_data_version'})) {
		SetErrMsg("MariaDB_SelfArgCheck: app_data_version: must be format %d.%d or 0", $errMsg);
		goto End;
	}

	$bkpDataVersion = floatval($bkpInput->{'app_data_version'});
	if ($DATA_VERSION < $bkpDataVersion) {
		SetErrMsg("MariaDB_SelfArgCheck: app_data_version: backup data version ($bkpDataVersion) must smaller or equal to package backup framework version ($DATA_VERSION)", $errMsg, ERR_MSG_KEY_RESTORE_DATA_TOO_NEW);
		goto End;
	}

	$ret = 0;
End:
	return $ret;
}

/**
 * Check function argument.
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 * @param bkpOutput [IN] Backup framework output argument, write output into this file path.
 *                    Should not be ""
 * @param type [IN] 0: export db, 1: import db
 * @param errMsg [OUT] Record error msg
 *
 * @return 0: success
 *        -1: fail
 */
function ArgumentCheck ($bkpInput, $bkpOutout, $type, &$errMsg)
{
	$ret = -1;

	if ("" === $bkpInput || "" === $bkpOutout) {
		SetErrMsg("MariaDB_ArgumentCheck: bkpInput/bkpOutout: must have value!", $errMsg);
		goto End;
	}

	if (!property_exists($bkpInput, 'data_range')) {
		SetErrMsg("MariaDB_ArgumentCheck: data_range: no param!", $errMsg);
		goto End;
	}

	if (!property_exists($bkpInput->{'data_range'}, 'this_app_data')) {
		SetErrMsg("MariaDB_ArgumentCheck: this_app_data: no param!", $errMsg);
		goto End;
	}

	if (!property_exists($bkpInput->{'data_range'}, 'other_app_data')) {
		SetErrMsg("MariaDB_ArgumentCheck: other_app_data: no param!", $errMsg);
		goto End;
	}

	// check self temp_path
	if ($bkpInput->{'data_range'}->{'this_app_data'}) {
		if (0 > SelfArgCheck($bkpInput, $bkpOutout, $type, $errMsg)) {
			goto End;
		}
	}

	$ret = 0;
End:
	return $ret;
}

/**
 * Get in-system database list
 *
 * @param dbs [OUT] Output array, put db list into this array
 *
 * @return 0: success
 *        -1: fail
 */
function DBListGet (&$dbs)
{
	$ret = -1;
	$qResult = null;

	exec("/usr/bin/mysql -e 'exit'",$execOutput, $execStatus);
	if (0 != $execStatus) {
		syslog(LOG_ERR, "MariaDB_DBListGet: Failed to connect mysql'");
		goto End;
	}

	if (!($qResult=popen("/usr/bin/mysql -sNe 'show databases;'", "r"))) {
		syslog(LOG_ERR, "MariaDB_DBListGet: Failed to execute show database query'");
		goto End;
	}
	while (($buf = fgets($qResult, 4096)) !== false) {
		array_push($dbs, trim($buf));
	}

	$ret = 0;
End:
	if ($qResult) {
		pclose($qResult);
	}
	return $ret;
}

/**
 * Export self(ALL) databases
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 *
 * @return 0: success
 *        -1: fail
 */
function SelfExport ($bkpInput, &$mysqldumpErrMsg)
{
	global $DangerWord;
	global $blCancel;
	$ret = -1;
	$tempPath = "";
	$databases = array();
	$execOutput = array();
	$execStatus = -1;

	$tempPath = escapeshellarg($bkpInput->{'temp_path'});

	if (0 > DBListGet($databases)) {
		syslog(LOG_ERR, "MariaDB_SelfExport: DBListGet: Failed to get database list");
		goto End;
	}

	foreach ($databases as $dbName) {
		if (true == $blCancel) {
			goto End;
		}

		if ("information_schema" == "$dbName" || "performance_schema" == "$dbName") {
			continue;
		}

		foreach ($DangerWord as $s) {
			if (false !== strstr($dbName, $s)) {
				syslog(LOG_ERR, "MariaDB_SelfExport: db name ($dbName) has danger word ($s)");
				goto End;
			}
		}

		$escName = escapeshellarg($dbName);
		//prevent database name contains '/' character
		$urlName = urlencode($dbName);
		$cmd = "/usr/bin/mysqldump --events --routine --trigger --single-transaction --max-allowed-packet=512M --database -- $escName > $tempPath/$urlName.sql";
		if (0 != exec_get_stderr($cmd, $stderr)) {
			syslog(LOG_ERR, "MariaDB_SelfExport: mysqldump: Failed to dump ($escName) to ($urlName.sql)");
			$mysqldumpErrMsg = "$stderr";
			goto End;
		}

		if (true == $blCancel) {
			goto End;
		}
	}

	$ret = 0;
End:
	return $ret;
}

/**
 * Do export process
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 *
 * @return 0: success
 *        -1: fail
 */
function DoExport ($bkpInput, &$mysqldumpErrMsg)
{
	$ret = -1;

	if ($bkpInput->{'data_range'}->{'this_app_data'}) {
		if (0 > SelfExport($bkpInput, $mysqldumpErrMsg)) {
			goto End;
		}
	}

	$ret = 0;
End:
	return $ret;
}

/**
 * Rename database from shellEscDbOld to shellEscDbNew if shellEscDbOld exist
 *
 * @param shellEscDbOld [IN] Original database name (shell escaped)
 * @param shellEscDbNew [IN] New database name (shell escaped)
 * @param ignoreCancel [IN] Skip $blCancel flag check
 *
 * @return 0: success
 *        -1: fail
 */
function DBReNameIfExist ($shellEscDbOld, $shellEscDbNew, $ignoreCancel = false)
{
	global $blCancel;
	$ret = -1;
	$execOutput = array();
	$execStatus = -1;

	if ("" === $shellEscDbOld || "" === $shellEscDbNew) {
		syslog(LOG_ERR, "MariaDB_DBReNameIfExist: old db ($shellEscDbOld)/new db ($shellEscDbNew) must have value!");
		goto End;
	}

	exec("/usr/bin/mysql -e \"use \\`\"$shellEscDbOld\"\\`;\"", $execOutput, $execStatus);
	if (0 != $execStatus) {
		$ret = 0;
		goto End;
	}

	exec("/usr/bin/mysql -e \"drop database if exists \\`\"$shellEscDbNew\"\\`;\"");
	exec("/usr/bin/mysql -e \"create database \\`\"$shellEscDbNew\"\\`;\"", $execOutput, $execStatus);
	if (0 != $execStatus) {
		syslog(LOG_ERR, "MariaDB_DBReNameIfExist: mysql: Failed to create database ($shellEscDbNew)!");
		goto End;
	}

	exec("/usr/bin/mysqldump --events --routine --triggers --max-allowed-packet=512M -- $shellEscDbOld | /usr/bin/mysql -- $shellEscDbNew", $execOutput, $execStatus);
	if (0 != $execStatus) {
		syslog(LOG_ERR, "MariaDB_DBReNameIfExist: mysqldump/mysql: Failed to rename db from ($shellEscDbOld) to ($shellEscDbNew)!");
		goto End;
	}

	if ((true == $blCancel) && !$ignoreCancel) {
		goto End;
	}

	exec("/usr/bin/mysql -e \"drop database if exists \\`\"$shellEscDbOld\"\\`;\"");
	$ret = 0;
End:
	if (0 > $ret) {
		exec("/usr/bin/mysql -e \"drop database if exists \\`\"$shellEscDbNew\"\\`;\"");
	}
	return $ret;
}

/**
 * Import self(ALL) databases
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 * @param overwrite [IN] true: always overwrite conflict databases
 *                      false: skip import conflict databases
 *
 * @return 0: success
 *        -1: fail
 */
function SelfImport ($bkpInput, $overwrite)
{
	global $DangerWord;
	global $blCancel;
	$ret = -1;
	$tempPath = "";
	$dbPaths = array();
	$execStatus = -1;
	$databases = array();
	$execOutput = array();
	$newDb = array();
	$randExt = mt_rand();

	$tempPath = escapeshellarg($bkpInput->{'temp_path'});
	exec("/usr/bin/find $tempPath -type f", $dbPaths, $execStatus);
	if (0 != $execStatus) {
		syslog(LOG_ERR, "MariaDB_SelfImport: find: Failed to enum file list ($tempPath)!");
		goto End;
	}

	if (0 > DBListGet($databases)) {
		syslog(LOG_ERR, "MariaDB_SelfImport: DBListGet: Failed to get database list");
		goto End;
	}

	foreach ($dbPaths as $urlPath) {
		if (true == $blCancel) {
			goto End;
		}

		// <MariaDB> #122 - In order to support old backup method, check file extension first
		$fext = pathinfo($urlPath, PATHINFO_EXTENSION);
		if ("gz" == $fext) {
			$dbName = urldecode(basename($urlPath, ".sql.gz"));
		} else if ("sql" == $fext) {
			$dbName = urldecode(basename($urlPath, ".sql"));
		} else {
			syslog(LOG_ERR, "Unsupported restore file type $fext ($urlPath)");
			goto End;
		}

		$shellEscDb = escapeshellarg($dbName);

		foreach ($DangerWord as $s) {
			if (false !== strstr($dbName, $s)) {
				syslog(LOG_ERR, "MariaDB_SelfImport: db name ($dbName) has danger word ($s)");
				goto End;
			}
		}

		//check if db already exist?
		if (in_array($dbName, $databases)) {
			if (!$overwrite) {
				continue;
			}
			//rename ori db table
			if (0 > DBReNameIfExist($shellEscDb, $shellEscDb."_$randExt")) {
				goto End;
			}
		}
		array_push($newDb, $shellEscDb);

		if (true == $blCancel) {
			goto End;
		}

		if ("gz" == $fext) {
			exec("/bin/gunzip -c $urlPath | /usr/bin/mysql", $execOutput, $execStatus);
		} else if ("sql" == $fext) {
			exec("/usr/bin/mysql < $urlPath", $execOutput, $execStatus);
		}
		if (0 != $execStatus) {
			syslog(LOG_ERR, "MariaDB_SelfImport: gunzip/mysql: Failed to decompress ($urlPath)");
			goto End;
		}
	}

	//drop all original databases
	foreach ($newDb as $shellEscDb) {
		$oriDb = $shellEscDb."_$randExt";
		exec("/usr/bin/mysql -e \"drop database if exists \\`\"$oriDb\"\\`;\"", $execOutput, $execStatus);
	}

	$ret = 0;
End:
	if (0 > $ret) {
		//drop all new databases
		foreach ($newDb as $shellEscDb) {
			exec("/usr/bin/mysql -e \"drop database if exists \\`\"$shellEscDb\"\\`;\"", $execOutput, $execStatus);
			DBReNameIfExist($shellEscDb."_$randExt", $shellEscDb, true);
		}
	}
	return $ret;
}

/**
 * Do import process
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 * @param overwrite [IN] Indicate if need to overwrite when DB conflict
 *
 * @return 0: success
 *        -1: fail
 */
function DoImport ($bkpInput, $overwrite)
{
	$ret = -1;

	if ($bkpInput->{'data_range'}->{'this_app_data'}) {
		if (0 > SelfImport($bkpInput, $overwrite)) {
			goto End;
		}
	}

	$ret = 0;
End:
	return $ret;
}

/**
 * Enter single user mode
 *
 * @return 0: success
 *        -1: fail
 */
function SingleUserStart ()
{
	$ret = -1;
	$execOutput = array();
	$execStatus = -1;

	exec("/var/packages/MariaDB/scripts/start-stop-status start singleuser >/dev/null 2>&1", $execOutput, $execStatus);
	if (0 != $execStatus) {
		syslog(LOG_ERR, "MariaDB_SingleUserStart: Failed to start!");
		goto End;
	}

	if (0 > MariaDBWaitForEnabled()) {
		syslog(LOG_ERR, "MariaDB_SingleUserStart: Timeout for waiting mysql to be ready!");
		goto End;
	}

	$ret = 0;
End:
	return $ret;
}

/**
 * Leave single user mode
 */
function SingleUserStop ()
{
	exec("/var/packages/MariaDB/scripts/start-stop-status stop >/dev/null 2>&1");
}

/**
 * Export/Import database.
 * After this process, need to resume to original mysql status
 *
 * @param bkpInput [IN] Backup framework input argunment
 *                   Should not be null
 * @param bkpOutput [IN] Backup framework output argument, write output into this file path.
 *                    Should not be ""
 * @param type [IN] 0: export db, 1: import db
 * @param overwrite [IN] Indicate if need to overwrite when DB conflict.
 *                       Used only when type is import
 *
 * @return 0: success
 *        -1: fail
 */
function DBOp ($bkpInput, $bkpOutput, $type, $overwrite = true)
{
	$ret = -1;
	$errMsg = new stdClass();
	$userInfo = array();
	$outputA = array();
	$execOutput = array();
	$execStatus = -1;

	if (0 > ArgumentCheck($bkpInput, $bkpOutput, $type, $errMsg)) {
		goto End;
	}

	if (IsMariaDBEnabled()) {
		SetErrMsg("MariaDB_DBOp: IsMariaDBEnabled: Mysql should been shutdown before export/import process!", $errMsg);
		goto End;
	}

	// Enter single user mode
	if (0 > SingleUserStart()) {
		SetErrMsg("", $errMsg, ERR_MSG_KEY_OP_FAIL);
		goto End;
	}

	switch ($type) {
		case 0:
			$mysqldumpErrMsg = "";
			if (0 > DoExport($bkpInput, $mysqldumpErrMsg)) {
				if ("" != $mysqldumpErrMsg) {
					$errMsg = $mysqldumpErrMsg;
				} else {
					SetErrMsg("", $errMsg, ERR_MSG_KEY_BKP_FAIL);
				}
				goto End;
			}
			break;
		case 1:
			if (0 > DoImport($bkpInput, $overwrite)) {
				SetErrMsg("", $errMsg, ERR_MSG_KEY_RESTORE_FAIL);
				goto End;
			}
			break;
		default:
			SetErrMsg("MariaDB_DBOp: Wrong param! Support exportdb/importdb only!", $errMsg);
			goto End;
	}

	if (0 == $type) {
		$outputA[RET_KEY_DATA_VERSION] = DATA_VERSION;
	}

	$ret = 0;
End:
	SingleUserStop();
	$outputA[RET_KEY_RESULT] = (0 == $ret);
	if (0 > $ret) {
		$outputA[RET_KEY_ERR_MSG] = $errMsg;
	}
	file_put_contents($bkpOutput, json_encode($outputA));
	return $ret;
}


if (4 > $argc) {
	echo "exportdb/importdb need at least 4 arguments\n";
	syslog(LOG_ERR, "MariaDB_ParamCheck: exportdb/importdb need at least 4 arguments");
	exit(1);
}

$RET = 0;
switch ($argv[1]) {
	case "exportdb":
		if (0 > DBOp(json_decode($argv[2]), $argv[3], 0)) {
			$RET = 1;
		}
		break;
	case "importdb":
		if (5 > $argc) {
			echo "importdb need 5 arguments\n";
			syslog(LOG_ERR, "MariaDB_ParamCheck: importdb need 5 arguments");
			$RET = 1;
			break;
		}
		if (0 > DBOp(json_decode($argv[2]), $argv[3], 1, ("1" === $argv[4]))) {
			$RET = 1;
		}
		break;
	default:
		$RET = 1;
}

echo "End of bkpcommon script! RET: $RET\n";
exit($RET);

?>
